///////////////////////////////////////////////////////////////////////////////
//                                                         
// File.cc
// -------
// File class implementation
//                                               
// Design and Implementation by Bjoern Lemke               
//
// (C)opyright 2000-2013 Bjoern Lemke
//
// Design and Implementation by Bjoern Lemke               
//                                                         
// IMPLEMENTATION MODULE
//
// Class: File
// 
// Description: All base operations on files
//
// Status: IMPLEMENTED
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _REENTRANT
#define _REENTRANT    /* basic 3-lines for threads */
#endif


// SYSTEM INCLUDES
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

// BASE INCLUDES
#include "Exception.h"
#include "Chain.h"
#include "File.h"
#include "Tokenizer.h"

// DEFINES

#define BUFSIZE 1024

File::File()
{
     _fd = 0;
     _mode = CLOSED;
}

File::File(char *s)
{
    _fd = 0;
    _filename = s;
    _mode = CLOSED;
}

File::File(const Chain&  str)
{
     _fd = 0;
     _filename = str;
     _mode = CLOSED;
}

File::~File()
{
     
}

bool File::exists()
{
    struct stat sb;
    int ret = ::stat(_filename, &sb);
    
    if ( ret == -1 ) 
    {
	if ( errno == ENOENT )
	    return false;
	else
	{
	    Chain msg = Chain("Cannot stat file ") +  _filename + Chain(" : ") + Chain(strerror(errno));
	    throw Exception(EXLOC, msg);
	    
	}
    }
    return true;
}	  


void File::open(Mode m)
{
     int oflag;
     
     switch (m)
     {
     case READ:
#ifdef HAVE_MINGW32
	 oflag = O_RDONLY | O_BINARY;
#else
	 oflag = O_RDONLY;
#endif
	 break;
     case WRITE:
#ifdef HAVE_MINGW32
	 oflag = O_WRONLY | O_CREAT | O_TRUNC | O_BINARY;
#else
	 oflag = O_WRONLY | O_CREAT | O_TRUNC;
#endif
	 break;
     case READWRITE:
#ifdef HAVE_MINGW32
	 oflag = O_RDWR | O_CREAT | O_BINARY;
#else
	 oflag = O_RDWR | O_CREAT;
#endif
	 break;
     case APPEND:
#ifdef HAVE_MINGW32
	 oflag = O_APPEND | O_WRONLY | O_CREAT | O_BINARY;
#else
	 oflag = O_APPEND | O_WRONLY | O_CREAT;
#endif
	 break;
     case CLOSED:
	 throw Exception(EXLOC, Chain("Invalid mode for open"));
     }
     
     if (( _fd = ::open(_filename, oflag, 00644)) == -1)
     {
	 Chain msg = "Cannot open file " +  _filename;
	 throw Exception(EXLOC, msg);
     }	  

     _mode = m;
     
     return;
}


void File::close()
{
     if (_mode != CLOSED)
     {
	  ::close (_fd);
	  _mode = CLOSED;
	  _fd = 0;
     }
}

void File::writeByte( char *buf, unsigned long bufsize )
{
     if ( _mode != WRITE && _mode != READWRITE && _mode != APPEND )
     {
	 Chain msg = Chain("File ") + _filename + Chain(" not open for write");
	 throw Exception(EXLOC, msg);
     }

     unsigned long numbyte = 0;

     /*
     if ( ( numbyte = ::write(_fd, buf, bufsize)) == -1 )
     {
	 Chain msg = Chain("Write system error for ") +  _filename + Chain(" : ") + Chain(strerror(errno));
	 throw Exception(EXLOC, msg);
     }
     */

     while ( numbyte < bufsize )
     {
	 unsigned long n;
	 if ( ( n = ::write(_fd, (char*)((unsigned long)buf + numbyte), bufsize - numbyte)) == -1 )
	 {
	     Chain msg = Chain("Write system error for ") +  _filename + Chain(" : ") + Chain(strerror(errno));
	     throw Exception(EXLOC, msg);
	 }
	 numbyte += n;

	 /*
	 if ( numbyte < bufsize )
	     cout << "File : Write not complete on first try ( " 
		  << numbyte << " bytes written of " 
		  << bufsize << " )" << endl;
	 */
     }
}

unsigned long File::readByte( char *buf, unsigned long bufsize )
{
     if ( _mode != READ && _mode != READWRITE )
     {
	 Chain msg = Chain("File ") + _filename + Chain(" not open for read");
	 throw Exception(EXLOC, msg );
     }

     unsigned long numbyte = 0;

     if ( (numbyte = ::read(_fd, buf, bufsize)) == -1 )
     {
	 Chain msg = Chain("Read system error for ") +  _filename + Chain(" : ") + Chain(strerror(errno));
	 throw Exception(EXLOC, msg);
     }

     return numbyte;
}

void File::writeChain(const Chain& str)
{
     if ( _mode != WRITE && _mode != READWRITE && _mode != APPEND )
     {
	 Chain msg = Chain("File ") + _filename + Chain(" not open for write");
	 throw Exception(EXLOC, msg);
     }
     if (str.length() > 0 )
     {
	 if ( ::write(_fd, (char*)str, str.length()-1) == -1 )
	 {
	     Chain msg = Chain("Write system error for ") +  _filename + Chain(" : ") + Chain(strerror(errno));
	     throw Exception(EXLOC, msg);
	 }
     }
     return;
}

int File::readLine(Chain& line, long maxLen)
{

    if (_mode != READ && _mode != READWRITE )
    {
	Chain msg = Chain("File ") + _filename + Chain(" not open for read");
	throw Exception(EXLOC, msg);
    }

    char* lineBuf = new char[maxLen+1];

    bool eol = false;

    int i = 0;

    while (! eol && i < maxLen )
    {
	if ( ::read(_fd, &lineBuf[i], 1) == 1)
	{
	    if (lineBuf[i] == '\n') 
		eol = true;
	    else
		i++;
	} 
	else 
	{
	    if ( i  == 0 ) 
		return 0;
	    else
		eol = true;	       
	}
    }
    
    if ( i == maxLen )
    {
	throw Exception(EXLOC, "Line length exceeded");
    }

    lineBuf[i] = '\0';
    
    line = Chain(lineBuf);

    return i+1;
}

void File::seek(unsigned long pos)
{
    if ( ::lseek(_fd, pos, SEEK_SET) == -1 )
    {
	Chain msg = Chain("Seek system error for ") +  _filename + Chain(" : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
    return;
}


void File::trunc(unsigned long size)
{
    if ( ::ftruncate(_fd, size) == -1 )
    {
	Chain msg = Chain("ftruncate system error for ") +  _filename + Chain(" : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
    return;
}

void File::remove()
{

    if ( ::unlink(_filename) == -1 )
    {
	Chain msg = Chain("Unlink system error for ") +  _filename + Chain(" : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
    return;
    
}

int File::Size() const
{

    struct stat sb;

    if ( ::fstat(_fd, &sb) == -1 )
    {
	Chain msg = Chain("Fstat system error for ") +  _filename + Chain(" : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
    
    return sb.st_size;
    
}

const Chain& File::getFileName() const
{
    return _filename;
}

Chain File::getShortName() const
{
    
#ifdef HAVE_MINGW32
    // not supported
    return Chain("N/A");
#else
    Tokenizer tok(_filename, Chain("/"));
    Chain part;
    while (tok.nextToken(part));
    return part;
#endif
}


File& File::operator = ( File& f)
{
    if ( _mode != CLOSED || f._mode != CLOSED )
    {
	throw Exception(EXLOC, "Closed instances are required for copy"); 
    }
    
    open(WRITE);
    f.open(READ);
        
    unsigned long numByte;
    char buffer[BUFSIZE];
    	
    while ((numByte = f.readByte(buffer, BUFSIZE)) > 0)
    {
	writeByte(buffer, numByte); 
    }
    return (*this);
}

File& File::operator += ( File& f)
{

    if ( _mode != CLOSED || f._mode != CLOSED )
    {
	throw Exception(EXLOC, "Closed instances are required for append"); 
    }
    
    open(APPEND);
    f.open(READ);
        
    unsigned long numByte;
    char buffer[BUFSIZE];
    
    while ((numByte = f.readByte(buffer, BUFSIZE)) > 0)
    {
	writeByte(buffer, numByte); 
    }

    return (*this);
}


File& File::operator << ( const Chain& str)
{
    writeChain(str);
    return (*this);
}

File& File::operator << ( unsigned long l)
{
    Chain s(l);
    writeChain(s);
    return (*this);
}

File& File::operator << ( int i)
{
    Chain s(i);
    writeChain(s);
    return (*this);
}

File& File::operator << ( char c)
{
    char buf[2];
    buf[0] = c;
    buf[1] = 0;

    writeChain(buf);
    return (*this);
}







